/*

*/
import scene.*;
import scene.primitives.*;
import pipeline.*;
import framebuffer.*;

import java.awt.Color;
import java.awt.event.*;

/**
                   y
                   |
                   |
                   | v[4]
                 1 +-----------------+ v[5]=(1,1,0)
                  /|                /|
                /  |              /  |
              /    |            /    |
            /      |          /      |
      v[7] +-----------------+ v[6]  |
           |       |         |       |                    y
           |       |         |       |                    |
           |       |         |       | v[1]               |
           |  v[0] +---------|-------+------> x           |
           |      /          |      /1                    |
           |    /            |    /                       +-----> x
           |  /              |  /                        /
           |/                |/                         /
         1 +-----------------+                         /
         / v[3]=(0,0,1)      v[2]=(1,0,1)            z
       /
     /
    z

   Render a single wireframe cube.
*/
public class InteractiveCube_v16
{
   private static boolean render_z_buffer = false;
   private static boolean doBackFaceCulling = true;
   private static boolean frontFacingIsCCW = true;

   private static double eyeX = 0.0;
   private static double eyeY = 0.0;
   private static double eyeZ = 2.0;

   private static double xTranslation = 0.0;
   private static double yTranslation = 0.0;
   private static double zTranslation = 0.0;
   private static double xRotation = 0.0;
   private static double yRotation = 0.0;
   private static double zRotation = 0.0;
   private static double scale = 1.0;

   private static double far    =  2.0;
   private static double near   =  1.0;
   private static double left   = -1.0;
   private static double right  =  1.0;
   private static double bottom = -1.0;
   private static double top    =  1.0;

   private static boolean useFOVY = false;
   private static double fovy = 90.0;

   public static void main(String[] args)
   {
      System.out.println("Use the 'd' key to toggle debugging information on and off.");
      System.out.println("Use the 'p' key to toggle between parallel and orthographic projection.");
      System.out.println("Use the x/X, y/Y, z/Z, keys to translate the cube along the x, y, z axes.");
      System.out.println("Use the u/U, v/V, w/W, keys to rotate the cube around the x, y, z axes.");
      System.out.println("Use the s/S keys to scale the size of the cube.");
      System.out.println("Use the 'c' key to change the random vertex colors.");
      System.out.println("Use the 'a' key to toggle antialiasing on and off.");
      System.out.println("Use the n/N keys to move the camera's near plane.");
      System.out.println("Use the b/B keys to move the camera's far (back) plane.");
      System.out.println("Use the o/O keys to zoooom in and out on the cube.");
      System.out.println("Use the arrow keys to move the camera location up/down and left/right.");
      System.out.println("Use the ' ' key to toggle between the image buffer and z-buffer.");
      System.out.println("Use the 'f' key to toggle fixing the perspective field of view.");
      System.out.println("Use the 'g' key to toggle back face culling.");
      System.out.println("Use the 'h' key to toggle CW vs CCW.");

      // Create a Model object to hold the geometry.
      final Model model = new Model();

      // Add the geometry (in model coordinates) to the Model.
      final Vertex[] v = new Vertex[8];

      v[0] = new Vertex(0.0, 0.0, 0.0);  // four vertices around the bottom face
      v[1] = new Vertex(1.0, 0.0, 0.0);
      v[2] = new Vertex(1.0, 0.0, 1.0);
      v[3] = new Vertex(0.0, 0.0, 1.0);
      v[4] = new Vertex(0.0, 1.0, 0.0);  // four vertices around the top face
      v[5] = new Vertex(1.0, 1.0, 0.0);
      v[6] = new Vertex(1.0, 1.0, 1.0);
      v[7] = new Vertex(0.0, 1.0, 1.0);

      v[0].setColorRandom();
      v[1].setColorRandom();
      v[2].setColorRandom();
      v[3].setColorRandom();
      v[4].setColorRandom();
      v[5].setColorRandom();
      v[6].setColorRandom();
      v[7].setColorRandom();

      model.addPrimitive( new Face(v[0], v[1], v[2], v[3]) );  // bottom face
      model.addPrimitive( new Face(v[4], v[7], v[6], v[5]) );  // top face
      model.addPrimitive( new Face(v[0], v[4], v[5], v[1]) );  // back face
      model.addPrimitive( new Face(v[3], v[2], v[6], v[7]) );  // front face
      model.addPrimitive( new Face(v[1], v[5], v[6], v[2]) );  // right face
      model.addPrimitive( new Face(v[0], v[3], v[7], v[4]) );  // left face
/*
      model.addPrimitive( new LineSegment(v[0], v[1]) );  // bottom face
      model.addPrimitive( new LineSegment(v[1], v[2]) );  // bottom face
      model.addPrimitive( new LineSegment(v[2], v[3]) );  // bottom face
      model.addPrimitive( new LineSegment(v[3], v[0]) );  // bottom face
      model.addPrimitive( new LineSegment(v[4], v[5]) );  // top face
      model.addPrimitive( new LineSegment(v[5], v[6]) );  // top face
      model.addPrimitive( new LineSegment(v[6], v[7]) );  // top face
      model.addPrimitive( new LineSegment(v[7], v[4]) );  // top face
      model.addPrimitive( new LineSegment(v[0], v[4]) );  // back face
      model.addPrimitive( new LineSegment(v[1], v[5]) );  // back face
      model.addPrimitive( new LineSegment(v[3], v[7]) );  // front face
      model.addPrimitive( new LineSegment(v[2], v[6]) );  // front face
*/

      // Create the Scene object that we shall render.
      final Scene scene = new Scene();

      // Add the Model to the Scene.
      scene.addModel(model);


      // Create a Camera object.
      Camera camera = new Camera();

      // Set up the camera's view volume.
      camera.projPerspective(left, right, bottom, top, near, far);
/*
      // An alternative way to set up the camera's view frustum.
      double fov    = 90.0;
      double aspect = 1.0;
      camera.projPerspective(fov, aspect, near, far);
*/
      // Set up the camera's location and orientation.
      double centerX = 0, centerY = 0, centerZ = 0;
      double     upX = 0,     upY = 1,     upZ = 0;
      camera.viewLookAt(eyeX,    eyeY,    eyeZ,
                        centerX, centerY, centerZ,
                        upX,     upY,     upZ);

      // Add the Camera to the Scene.
      scene.setCamera(camera);


      // Create a InteractiveFrame and a FrameBuffer to render our scene into.
      int width  = 512;
      int height = 512;

      @SuppressWarnings("serial")
      final InteractiveFrame app = new InteractiveFrame("Interactive JFrame", width, height)
      {
         // Implement part of the KeyListener interface.
         @Override public void keyPressed(KeyEvent e)
         {
            //System.out.println( e );
            int keyCode = e.getKeyCode();
            if (KeyEvent.VK_UP == keyCode)
            {
               eyeY += 0.1;
            }
            else if (KeyEvent.VK_DOWN == keyCode)
            {
               eyeY -= 0.1;
            }
            else if (KeyEvent.VK_LEFT == keyCode)
            {
               eyeX -= 0.1;
            }
            else if (KeyEvent.VK_RIGHT == keyCode)
            {
               eyeX += 0.1;
            }

            // Set up the camera's location and orientation.
            scene.camera.view2Identity();
            double centerX = 0, centerY = 0, centerZ = 0;
            double upX = 0,     upY = 1,     upZ = 0;
            scene.camera.viewLookAt(eyeX,    eyeY,    eyeZ,
                                    centerX, centerY, centerZ,
                                    upX,     upY,     upZ);

            // Render again.
            this.getFrameBuffer().clearFB(Color.black);
            Pipeline.render(scene, this.getFrameBuffer());
            repaint();
         }

         // Implement part of the KeyListener interface.
         @Override public void keyTyped(KeyEvent e)
         {
            //System.out.println( e );
            char c = e.getKeyChar();
            if ('d' == c)
            {
             //Pipeline.debug = ! Pipeline.debug;
               Pipeline.debug_visible = ! Pipeline.debug_visible;
             //Rasterize.debug = ! Rasterize.debug;
             //RasterizeAntialias.debug = ! RasterizeAntialias.debug;
            }
            else if ('g' == c)
            {
               doBackFaceCulling = ! doBackFaceCulling;
               for (Model m : scene.modelList)
                  m.doBackFaceCulling = doBackFaceCulling;
               System.out.print("Back face culling is ");
               System.out.println(doBackFaceCulling ? "on" : "off");
            }
            else if ('h' == c)
            {
               frontFacingIsCCW = ! frontFacingIsCCW;
               for (Model m : scene.modelList)
                  m.frontFacingIsCCW = frontFacingIsCCW;
               System.out.print("Front face is ");
               System.out.println(frontFacingIsCCW ? "CCW" : "CW");
            }
            else if ('f' == c)
            {
               useFOVY = ! useFOVY;
               if ( useFOVY )
               {
                  fovy = 2.0 * (180./Math.PI) * Math.atan(top/near);
                  System.out.println("The field of view is fixed at " + fovy + ".");
               }
               else
               {
                  top    =  near * Math.tan((Math.PI/180.0)*fovy/2.0);
                  bottom = -top;
                  right  =  top;
                  left   = -right;
                  System.out.println("The field of view will change with near plane.");
               }
            }
            else if ('o' == c || 'O' == c)
            {
               if ( useFOVY )
               {
                  // Update the fovy.
                  if ('o' == c)
                  {
                     fovy -= 1.0;
                  }
                  else
                  {
                     fovy += 1.0;
                  }
               }
               else
               {
                  // Change left, right, bottom, and top.
                  if ('o' == c)
                  {
                     left   += 0.1;
                     right  -= 0.1;
                     bottom += 0.1;
                     top    -= 0.1;
                  }
                  else
                  {
                     left   -= 0.1;
                     right  += 0.1;
                     bottom -= 0.1;
                     top    += 0.1;
                  }
               }
            }
            else if ('n' == c || 'N' == c)
            {
               // Move the near plane.
               if ('n' == c)
                  near += 0.01;
               else
                  near -= 0.01;
            }
            else if ('b' == c || 'B' == c)
            {
               // Move the far (back) plane.
               if ('b' == c)
                  far += 0.01;
               else
                  far -= 0.01;
            }
            else if (' ' == c)
            {
               render_z_buffer = ! render_z_buffer;
            }
            else if ('a' == c)
            {
               Pipeline.doAntialiasing = ! Pipeline.doAntialiasing;
            }
            else if ('c' == c)
            {
               // Change the random color at each vertex.
               for (int i = 0; i < v.length; i++)
               {
                  v[i].setColorRandom();
               }
            }
            else if ('p' == c)
            {
               scene.camera.perspective = ! scene.camera.perspective;
            }
            else if ('s' == c) // Scale the model 10% smaller.
            {
               scale *= 0.9;
            }
            else if ('S' == c) // Scale the model 10% larger.
            {
               scale *= 1.1;
            }
            else if ('x' == c)
            {
               xTranslation += 0.1;
            }
            else if ('X' == c)
            {
               xTranslation -= 0.1;
            }
            else if ('y' == c)
            {
               yTranslation += 0.1;
            }
            else if ('Y' == c)
            {
               yTranslation -= 0.1;
            }
            else if ('z' == c)
            {
               zTranslation += 0.1;
            }
            else if ('Z' == c)
            {
               zTranslation -= 0.1;
            }
            else if ('u' == c)
            {
               xRotation += 2.0;
            }
            else if ('U' == c)
            {
               xRotation -= 2.0;
            }
            else if ('v' == c)
            {
               yRotation += 2.0;
            }
            else if ('V' == c)
            {
               yRotation -= 2.0;
            }
            else if ('w' == c)
            {
               zRotation += 2.0;
            }
            else if ('W' == c)
            {
               zRotation -= 2.0;
            }

            // Set the model-to-world transformation matrix.
            // The order of the transformations is very important!
            model.model2Identity();
            model.modelTranslate(xTranslation, yTranslation, zTranslation);
            model.modelRotate(xRotation, 1, 0, 0);
            model.modelRotate(yRotation, 0, 1, 0);
            model.modelRotate(zRotation, 0, 0, 1);
            model.modelScale(scale, scale, scale);

            // Set up the camera's view volume.
            if (scene.camera.perspective)
            {
               if ( ! useFOVY )
               {
                  scene.camera.projPerspective(left, right, bottom, top, near, far);
               }
               else
               {
                  // An alternative way to set up the camera's view frustum.
                  scene.camera.projPerspective(fovy, 1, near, far);
               }
            }
            else
            {
               scene.camera.projOrtho(left, right, bottom, top, near, far);
            }
            if ('n'==c||'N'==c||'b'==c||'B'==c||'o'==c||'O'==c)
               System.out.print( scene.camera );

            // Render again.
            this.getFrameBuffer().clearFB(Color.black);
            Pipeline.render(scene, this.getFrameBuffer());
            if ( render_z_buffer )
            {
               FrameBuffer zBuf = this.getFrameBuffer().convertZB2FB();
               this.getFrameBuffer().setViewport(0, 0, zBuf);
            }
            repaint();
         }

         // Implement part of the ComponentListener interface.
         @Override public void componentResized(ComponentEvent e)
         {
            //System.out.println( e );
            java.awt.Container c = (java.awt.Container)e.getSource();
            java.awt.Insets insets = c.getInsets();
            int w = c.getWidth()  - insets.left - insets.right;
            int h = c.getHeight() - insets.top  - insets.bottom;

            this.createFrameBuffer(w, h);
            this.getFrameBuffer().clearFB(Color.black);
            Pipeline.render(scene, this.getFrameBuffer());
            repaint();
         }
      };

      app.addKeyListener(app);
      app.addComponentListener(app);
   }
}
