/*

*/
package pipeline;
import scene.*;
import scene.primitives.*;

/**
   Cull from the list of primitives any Face that is "back facing".
<p>
   For a good explanation of orientation, forward and backward facing polygons,
   and back face culling, see Chapter 9 of<br>
      "Computer Graphics Through OpenGL: From Theory to Experiments," 2nd Edition,
       by Sumanta Guha, Chapman and Hall/CRC, 2014.
<p>
   That chapter is available at<br>
       http://www.sumantaguha.com/files/materials/ch9.pdf
*/
public class BackFaceCulling
{
   /**
      Determine if a Face is forward or backward
      facing. The renderer will delete any Face
      that is backward facing.
   */
   public static boolean isBackFace(Primitive p, boolean frontFacingIsCCW)
   {
      return (p instanceof Face) && (! faceIsFrontFacing((Face)p, frontFacingIsCCW));
   }


   private static boolean faceIsFrontFacing(Face face, boolean frontFacingIsCCW)
   {
      boolean result = faceIsCCW(face);
      if ( ! frontFacingIsCCW )
         result = ! result;
      return result;
   }


   /**
      In view coordinates, this version is accurate when using
      an orthographic projection but it is only approximately
      correct when using a perspective projection.

      Also, this version can be used in Normalized Device Coordinates
      (after the projection stage).
   */
   private static boolean faceIsCCW_(Face face)
   {
      // We only need to know the z component of the cross product
      // of the two "edge vectors", v1-v0 and v2-v0. This works
      // because we are in view coordinates, with the camera at
      // the origin looking down the negative z-axis.
      double x0 = face.vertexList.get(0).x;
      double x1 = face.vertexList.get(1).x;
      double x2 = face.vertexList.get(2).x;
      double y0 = face.vertexList.get(0).y;
      double y1 = face.vertexList.get(1).y;
      double y2 = face.vertexList.get(2).y;
      double zValueOfCrossProduct = (x1-x0)*(y2-y0) - (y1-y0)*(x2-x0);

      //System.err.printf("zValueOfCrossProduct = % .6f\n", zValueOfCrossProduct);

      return (zValueOfCrossProduct >= 0);
   }


   /**
      In view coordinates, this version is accurate when
      using a perspective projection.
   */
   private static boolean faceIsCCW(Face face)
   {
      double x0 = face.vertexList.get(0).x;
      double x1 = face.vertexList.get(1).x;
      double x2 = face.vertexList.get(2).x;
      double y0 = face.vertexList.get(0).y;
      double y1 = face.vertexList.get(1).y;
      double y2 = face.vertexList.get(2).y;
      double z0 = face.vertexList.get(0).z;
      double z1 = face.vertexList.get(1).z;
      double z2 = face.vertexList.get(2).z;
      double determinant = x0*(y1*z2-y2*z1)-x1*(y0*z2-y2*z0)+x2*(y0*z1-y1*z0);

      //System.err.printf("determinant = % .6f\n", determinant);

      return (determinant < 0);
   }
}
