/*

*/
package pipeline;
import scene.*;
import scene.primitives.LineSegment;
import framebuffer.*;

import java.util.List;
import java.util.ListIterator;

/**
    This renderer takes as its input a Scene data structure and
    a FrameBuffer data structure. This renderer mutates the
    FrameBuffer's current viewport so that it is filled in with
    the rendered image of the scene represented by the Scene object.
<p>
    This implements our sixteenth rendering pipeline. This renderer
    adds a new geometric primitive, an oriented face, and a new
    pipeline stage, backface culling, that deletes from the rendering
    pipeline any oriented face that is not facing the camera.
*/
public class Pipeline
{
   public static boolean doBackFaceCulling;
   public static boolean frontFacingIsCCW;
   public static boolean doAntialiasing = false;
   public static boolean debug = false;
   public static boolean debug_visible = false;

   /**
      Mutate the FrameBuffer's current viewport so that it holds
      the rendered image of the Scene object.

      @param scene Scene object to render
      @param fb    FrameBuffer to hold rendered image of the scene
   */
   public static void render(Scene scene, FrameBuffer fb)
   {
      if ( debug || debug_visible )
         render_debug(scene, fb);
      else
         render_(scene, fb);
   }


   /**
      Renderer a scene without any debugging information.

      @param scene Scene object to render
      @param fb    FrameBuffer to hold rendered image of the scene
   */
   private static void render_(Scene scene, FrameBuffer fb)
   {
      // For every Model in the Scene, render the model
      // and every sub model in the Model.
      for (Model model : scene.modelList)
      {
         doBackFaceCulling = model.doBackFaceCulling;
         frontFacingIsCCW  = model.frontFacingIsCCW;

         render_model(scene, model, scene.camera.viewMatrix, fb);
      }
   }


   /**
      Recursively renderer a model without any debugging information.

      @param scene the Scene that we are rendering
      @param model the current Model object to recursively render
      @param ctm   current transformation matrix
      @param fb    FrameBuffer to hold rendered image of the model
   */
   private static void render_model(Scene scene, Model model, Matrix ctm, FrameBuffer fb)
   {
      // Get the viewport dimensions (needed for the viewport stage).
      int w = fb.getWidthVP();
      int h = fb.getHeightVP();

      // Update the current transformation matrix.
      ctm = ctm.times( model.modelMatrix );

      // Render every Primitive in the primitive list from the Model.
      for (Primitive p : model.primitiveList)
      {
         // Make a deep copy of the primitive.
         Primitive p2 = p.makeCopy();

         // 1. Apply the current model-to-world, world-to-view
         //    transformation to each vertex in the primitive
         for (Vertex v : p2.vertexList)
         {
            v.times(ctm);  // mutate the Vertex object
         }

         // 2. Do the back face culling operation.
         if ( ! doBackFaceCulling
           || ! BackFaceCulling.isBackFace(p2, frontFacingIsCCW) )
         {
            // Create the combined normalize, projection transformation.
            Matrix np = scene.camera.projMatrix.times(
                           scene.camera.normalizeMatrix);

            // 3. Apply the normalize, projection transformation
            //    to each vertex in the primitive.
            for (Vertex v : p2.vertexList)
            {
               v.times(np);
            }

            // 4. Assemble the primitive into a list of line segments.
            List<LineSegment> lineSegmentList = p2.assemble();

            // Process all the line segments from this primitive.
            for (LineSegment ls : lineSegmentList)
            {
               // 5. Clip the line segment to the (homogeneous) standard clip region.
               if ( Clip.clip(ls) )
               {
                  Vertex v0 = ls.vertexList.get(0);
                  Vertex v1 = ls.vertexList.get(1);

                  // 6. Apply the perspective division operation (clip-to-NDC coordinates).
                  v0.perspectiveDivision();
                  v1.perspectiveDivision();

                  // 7. Apply the viewport transformation (NDC-to-viewport coordinates).
                  v0.x = 0.5 + w/2.001 * (v0.x + 1.0);
                  v0.y = 0.5 + h/2.001 * (v0.y + 1.0);
                  v1.x = 0.5 + w/2.001 * (v1.x + 1.0);
                  v1.y = 0.5 + h/2.001 * (v1.y + 1.0);

                  // 8. Rasterize the line segment into pixels.
                  if ( doAntialiasing )
                  {
                     RasterizeAntialias.rasterize(ls, fb);
                  }
                  else
                  {
                     Rasterize.rasterize(ls, fb);
                  }
               }
            }
         }
      }

      // Recursively render every sub model of this model.
      for (Model m : model.nestedModels)
      {
         render_model(scene, m, ctm, fb);
      }
   }



   /**
      Render a scene with quite a bit of debugging information.

      @param scene Scene object to render
      @param fb    FrameBuffer to hold rendered image of the scene
   */
   private static void render_debug(Scene scene, FrameBuffer fb)
   {
      System.out.println("=== Render Scene ===================================");

      // For every Model in the Scene, render the model
      // and every sub model in the Model.
      for (Model model : scene.modelList)
      {
         System.out.println("=== Render Model ===================================");
         render_model_debug(scene, model, scene.camera.viewMatrix, fb);
         System.out.println("=== End Model ======================================");
      }

      System.out.println("=== End Scene ===================================");
   }


   /**
      Recursively renderer a model with quite a bit of debugging information.

      @param scene the Scene that we are rendering
      @param model the current Model object to recursively render
      @param ctm   current transformation matrix
      @param fb    FrameBuffer to hold rendered image of the model
   */
   private static void render_model_debug(Scene scene, Model model, Matrix ctm, FrameBuffer fb)
   {
      // Get the viewport dimensions (needed for the viewport stage).
      int w = fb.getWidthVP();
      int h = fb.getHeightVP();

      // Update the current transformation matrix.
      ctm = ctm.times( model.modelMatrix );

      // Render every Primitive in the primitive list from the Model.
      for (Primitive p : model.primitiveList)
      {
         StringBuilder result = new StringBuilder();

         // Make a deep copy of the primitive.
         Primitive p2 = p.makeCopy();

         // log interesting information to standard output
         logPrimitive("0. Model", p2, result);

         // 1. Apply the current model-to-world, world-to-view
         //    transformation to each vertex in the primitive
         for (Vertex v : p2.vertexList)
         {
            v.times(ctm);  // mutate the Vertex object
         }

         // log interesting information to standard output
         logPrimitive("1. View", p2, result);

         // 2. Do the back face culling operation.
         if ( ! model.doBackFaceCulling
           || ! BackFaceCulling.isBackFace(p2, model.frontFacingIsCCW) )
         {
            // Create the combined normalize, projection transformation.
            Matrix np = scene.camera.projMatrix.times(
                           scene.camera.normalizeMatrix);

            // 3. Apply the normalize, projection transformation
            //    to each vertex in the primitive.
            for (Vertex v : p2.vertexList)
            {
               v.times(np);
            }

            // log interesting information to standard output
            logPrimitive("3. Clip", p2, result);

            // 4. Assemble the primitive into a list of line segments.
            List<LineSegment> lineSegmentList = p2.assemble();

            // log interesting information to standard output
            for (LineSegment ls : lineSegmentList)
               logLineSegment("4. Assembled", ls, result);

            // Process all the line segments from this primitive.
            for (LineSegment ls : lineSegmentList)
            {
               // 5. Clip the line segment to the (homogeneous) standard clip region.
               if ( Clip.clip(ls) )
               {
                  // log interesting information to standard output
                  logLineSegment("5. Clipped", ls, result);

                  Vertex v0 = ls.vertexList.get(0);
                  Vertex v1 = ls.vertexList.get(1);

                  // 6. Apply the perspective division operation (clip-to-NDC coordinates).
                  v0.perspectiveDivision();
                  v1.perspectiveDivision();

                  // log interesting information to standard output
                  logLineSegment("6. Divided", ls, result);

                  // 7. Apply the viewport transformation (NDC-to-viewport coordinates).
                  v0.x = 0.5 + w/2.001 * (v0.x + 1.0);
                  v0.y = 0.5 + h/2.001 * (v0.y + 1.0);
                  v1.x = 0.5 + w/2.001 * (v1.x + 1.0);
                  v1.y = 0.5 + h/2.001 * (v1.y + 1.0);

                  // log interesting information to standard output
                  logLineSegment("7. Viewport", ls, result);

                  // 8. Rasterize the line segment into pixels.
                  if ( doAntialiasing )
                  {
                     RasterizeAntialias.rasterize(ls, fb);
                  }
                  else
                  {
                     Rasterize.rasterize(ls, fb);
                  }
                  if ( debug_visible ) System.out.println( result );
               }
               if ( ! debug_visible ) System.out.println();
            }
         }
      }

      // Recursively renderer every sub model in this model.
      for (Model m : model.nestedModels)
      {
         System.out.println("====== Render Sub Model ========================");
         render_model_debug(scene, m, ctm, fb);
         System.out.println("====== End Sub Model ===========================");
      }
   }


   /**
      This method either outputs a String representation of the given
      Primitive or it accumulates the String representation in the
      accumulator variable (for outputting later).

      @param stage name for the pipeline stage
      @param p Primitive whose string representation is to be appended to the accumulator
      @param accumulator accumulates the string representations for all the primitives in this stage
   */
   private static void logPrimitive(String stage, Primitive p, StringBuilder accumulator)
   {
      if ( debug_visible )
         accumulator.append( stage + " " + p.toString() );
      else
         System.out.print( stage + " " + p.toString() );
   }


   /**
      This method either outputs a String representation of the given
      LineSegment or it accumulates the String representation in the
      accumulator variable (for outputting later).

      @param stage name for the pipeline stage
      @param ls LineSegment whose string representation is to be appended to the accumulator
      @param accumulator accumulates the string representations for all the line segments in this stage
   */
   private static void logLineSegment(String stage, LineSegment ls, StringBuilder accumulator)
   {
      if ( debug_visible )
         accumulator.append( stage + " " + ls.toString() );
      else
         System.out.print( stage + " " + ls.toString() );
   }
}
