/*

*/
package scene;

/**
   This Camera data structure represents a camera that can be
   moved around and aimed within world coordinates.

   This camera has associated to it a location and an orientation
   in the world coordinate system. A Camera object is positioned
   and aimed within world coordinates using its viewLookAt() method,
   which determines the camera's view matrix.

   This camera also has associated to it a "view volume" that
   determines what part of space the camera "sees" in front of
   it when we use the camera to take a picture (that is, when
   we render a scene).

   This camera can "take a picture" two ways, using a
   perspective projection or using a parallel (orthographic)
   projection. Each way of taking a picture has a different
   shape for its view volume. The data in this data structure
   determines each of the two view volume shapes.

   For the perspective projection, the view volume (in view
   coordinates!) is an infinitely long pyramid that is formed
   by the pyramid with its apex at the origin and its base in
   the plane z = -near with edges x = left, x = right, y = top,
   and y = bottom.

   For the orthographic projection, the view volume (in view
   coordinates!) is an infinitely long rectangular cylinder
   parallel to the z-axis and with sides x = left, x = right,
   y = top, and y = bottom (an infinite parallelepiped).

   The part of space determined by the view volume is what
   the camera "sees" when we "take a picture" (that is,
   when we render a scene).

   The plane z = -near (in view coordinates) is the camera's view plane.
   The rectangle in the view plane with corners (left, bottom, -near)
   and (right, top, -near) is the camera's view rectangle. The view
   rectangle is like the film in a real camera, it is where the camera's
   image appears when you take a picture. The contents of the camera's
   view rectangle (after it gets "normalized") is what gets rasterized
   into the framebuffer.

   The projection matrix in this camera preserves some depth
   information. That is, this "projection" does not squash the
   camera's solid three-dimensional view volume into a flat
   two-dimensional rectangle.

   In fact, the perspective projection matrix in this class is a
   "homogeneous projection" that sends the camera's three-dimensional
   perspective view volume into a four-dimensional view region in
   four-dimensional homogeneous space. The perspective division
   operation will send that four-dimensional view region back to a
   three-dimensional view volume (in normalized device coordinates).

   The orthographic projection matrix in this class preserves depth
   information by simply being a translation matrix, so the solid
   three-dimensional orthographic view volume is just translated, and
   the perspective division operation has no affect on the view volume.

   In addition to preserving depth information, the projection matrix
   in this camera transforms the camera's near and far planes into the
   standard homogeneous front and back clipping planes, w - z = 0 and
   w + z = 0. This turns the camera's near and far planes into clipping
   planes. This camera's perspective view volume in now a frustum, and
   its orthographic view volume is a parallelepiped.
*/
public class Camera
{
   public boolean perspective;  // choose either perspective or parallel projection

   public Matrix viewMatrix;
   public Matrix normalizeMatrix;
   public Matrix projMatrix;

   public double left;   // These numbers define the camera's view volume.
   public double right;  // These numbers are encoded into the normalization matrix
   public double bottom;
   public double top;
   public double n;
   public double f;


   /**
      The default camera has the standard perspective view volume
      and is located at the origin (of world coordinates), looking
      down the negative z-axis.
   */
   public Camera()
   {
      this.projPerspective(-1.0, 1.0, -1.0, 1.0, 1.0, 2.0);

      viewMatrix = new Matrix(1.0); // identity matrix
   }


   /**
      Set up this camera's view volume as a perspective projection
      of an infinite view pyramid.

      @param left    left edge of view rectangle in the plane z = -near
      @param right   right edge of view rectangle in the plane z = -near
      @param bottom  bottom edge of view rectangle in the plane z = -near
      @param top     top edge of view rectangle in the plane z = -near
      @param near    distance from the origin to the near plane
      @param far     distance from the origin to the far plane
   */
   public void projPerspective(double left, double right, double bottom, double top, double near, double far)
   {
      this.left   = left;    // These numbers define the camera's view volume.
      this.right  = right;   // These numbers are encoded into the normalization matrix
      this.bottom = bottom;
      this.top    = top;
      this.n      = -near;
      this.f      = -far;

      this.projMatrix = new PerspectiveProjectionMatrix0(near, far);
      this.normalizeMatrix = new PerspectiveNormalizeMatrix(left, right, bottom, top, near);

      this.perspective = true;
   }


   /**
      An alternative way to determine the camera's perspective view volume.

      Here, the view volume is determined by a vertical "field of view"
      angle and an aspect ratio for the view rectangle.

      @param fovy    angle in the y-direction subtended by the view rectangle
      @param aspect  aspect ratio of the view rectangle
      @param near    distance from the origin to the near plane
      @param far     distance from the origin to the far plane
   */
   public void projPerspective(double fovy, double aspect, double near, double far)
   {
      this.top    =  near * Math.tan((Math.PI/180.0)*fovy/2.0);
      this.bottom = -top;
      this.right  =  top * aspect;
      this.left   = -right;
      this.n      = -near;
      this.f      = -far;

      projPerspective(left, right, bottom, top, near, far);
   }


   /**
      Set up this camera's view volume as a parallel (orthographic) projection
      of an infinite view parallelepiped.

      @param left    left edge of view rectangle in the plane z = -near
      @param right   right edge of view rectangle in the plane z = -near
      @param bottom  bottom edge of view rectangle in the plane z = -near
      @param top     top edge of view rectangle in the plane z = -near
      @param near    distance from the origin to the near plane
      @param far     distance from the origin to the far plane
   */
   public void projOrtho(double left, double right, double bottom, double top, double near, double far)
   {
      this.left   = left;    // These numbers define the camera's view volume.
      this.right  = right;   // These numbers are encoded into the normalization matrix
      this.bottom = bottom;
      this.top    = top;
      this.n      = 0;

      this.projMatrix = new OrthographicProjectionMatrix1(near, far);
      this.normalizeMatrix = new OrthographicNormalizeMatrix(left, right, bottom, top);

      this.perspective = false;
   }


   /**
      Set the location and orientation of this camera in the world coordinate system.

      Compare with
         https://www.opengl.org/sdk/docs/man2/xhtml/gluLookAt.xml

      @param eyex    x-coordinate of the camera's location
      @param eyey    y-coordinate of the camera's location
      @param eyez    z-coordinate of the camera's location
      @param centerx x-coordinate of the camera's look-at point
      @param centery y-coordinate of the camera's look-at point
      @param centerz z-coordinate of the camera's look-at point
      @param upx     x-component of the camera's up vector
      @param upy     y-component of the camera's up vector
      @param upz     z-component of the camera's up vector
   */
   public void viewLookAt(double eyex,    double eyey,    double eyez,
                          double centerx, double centery, double centerz,
                          double upx,     double upy,     double upz)
   {
      Vector F  = new Vector(centerx - eyex, centery - eyey, centerz - eyez);
      Vector UP = new Vector(upx, upy, upz);
      Vector f  = F.normalize();
      Vector up = UP.normalize();
      Vector s  = f.crossProduct(up);
      Vector u  = s.crossProduct(f);
      viewMatrix = new Matrix(
                       new Vector(s.x, u.x, -f.x, 0.0),
                       new Vector(s.y, u.y, -f.y, 0.0),
                       new Vector(s.z, u.z, -f.z, 0.0),
                       new Vector(0.0, 0.0,  0.0, 1.0));

      viewMatrix = viewMatrix.times(new Matrix(new Vector(-eyex, -eyey, -eyez)));
   }


   /**
      Set this camera's view matrix to the identity matrix.
   */
   public void view2Identity()
   {
      viewMatrix = new Matrix(1.0);
   }


   /**
      Multiply this camera's view matrix with the given matrix.

      @param m Matrix to multiply with this camera's view matrix
   */
   public void viewMult(Matrix m)
   {
      viewMatrix = viewMatrix.times( m );
   }


   /**
      Multiply this camera's view matrix with a rotation matrix.

      @param theta angle, in degrees, to rotate by
      @param x x-component of the axis vector for the rotation
      @param y y-component of the axis vector for the rotation
      @param z z-component of the axis vector for the rotation
   */
   public void viewRotate(double theta, double x, double y, double z)
   {
      viewMatrix = viewMatrix.times(new Matrix(theta, x, y, z));
   }


   /**
      Multiply this camera's view matrix with a translation matrix.

      @param x x-component of the translation vector
      @param y y-component of the translation vector
      @param z z-component of the translation vector
   */
   public void viewTranslate(double x, double y, double z)
   {
      viewMatrix = viewMatrix.times(new Matrix( new Vector(x, y, z)));
   }


   /**
      For debugging.

      @return String representation of this Camera object
   */
   public String toString()
   {
      double fovy = 2.0 * (180./Math.PI) * Math.atan(top/(-n));
      String result = "";
      result += "Camera: \n";
      result += "perspective = " + perspective + "\n";
      result += "left = "   + left + ", "
             +  "right = "  + right + "\n"
             +  "bottom = " + bottom + ", "
             +  "top = "    + top + "\n"
             +  "near = "   + -n + "\n"
             +  "far = "    + -f + "\n"
             +  "(fovy = " + fovy + ", aspect = " + right/top + ")\n"
             +  projMatrix + "\n"
             +  viewMatrix;
      return result;
   }
}
