/*

*/
package scene;

/**
<pre>
   This is a (non-singular) orthographic projection matrix.
     [ 1  0  0  0 ]
     [ 0  1  0  0 ]
     [ 0  0  1  1 ]
     [ 0  0  0  1 ]
   Notice that this matrix maps
        (x, y, z, 1) -> (x, y, z+1, 1).
   This matrix transforms the standard view parallelepiped
   (in camera coordinates) into the standard view parallelepiped
   (in clip coordinates, which are also NDC in this case) and it
   translates the standard view plane from z = -1 (in camera
   coordinates) to z = 0 (in NDC).

   But we want the near plane z = -near in camera coordinates to
   be transformed to the clipping plane at z = +1 in NDC and we want
   the far plane z = -far in camera coordinates to be transformed to
   the clipping plane at z = -1 in NDC So let us change the translation
   factor in the orthographic projection matrix,
     [ 1  0  0  0 ]
     [ 0  1  0  0 ]
     [ 0  0  1  b ]
     [ 0  0  0  1 ]
   so that the plane half way between near and far, z = -(near+far)/2
   in camera coordinates, gets transformed to z = 0. The new matrix
   maps
        (x, y, -(near+far)/2, 1) -> (x, y, -(near+far)/2 + b, 1)
   and we want
              -(near+far)/2 + b = 0
   so
               b = (near+far)/2.

   Now we need to follow the translation (in the z-direction) with a
   scaling (in the z-direction) so that z = -near gets transformed to
   z = +1 and z = -far gets transformed to z = -1. Solve for a in
     [ 1  0  0   0 ]
     [ 0  1  0   0 ]
     [ 0  0  a  a*b]
     [ 0  0  0   1 ]
   so that z = -near gets transformed to z = +1. This matrix maps
        (x, y, -near, 1) -> (x, y, -a*near + a*b, 1)
   and we want
            -a * near + a * b = 1
             a * (-near + b) = 1
             a * (-near + (near+far)/2) = 1
             a * (far - near)/2 = 1
             a = 2/(far - near)

   (You should check that z = -far transforms to z = +1.)
</pre>
*/
public class OrthographicProjectionMatrix1 extends Matrix
{
   /**
   <pre>
      Construct this orthographic projection matrix.
        [ 1  0  0  0 ]
        [ 0  1  0  0 ]
        [ 0  0  a  b ]
        [ 0  0  0  1 ]
      where
        a = 2/(far - near)
        b = (far + near)/(far - near).
   </pre>
   */
   public OrthographicProjectionMatrix1(double n, double f)
   {
      super(  new Vector(1.0, 0.0,    0.0,      0.0),
              new Vector(0.0, 1.0,    0.0,      0.0),
              new Vector(0.0, 0.0,     2/(f-n), 0.0),
              new Vector(0.0, 0.0, (f+n)/(f-n), 1.0) );
   }
}
