/*

*/
package scene;

/**
<pre>
   We use two steps to transform the camera's perspective view volume
   into the standard perspective view volume. The first step "skews" the
   current view volume so that its center line is the negative z-axis.
   The second step scales the skewed view volume so that it intersects
   the plane z = -1 with corners (-1, -1, -1) and (+1, +1, -1).

   This matrix skews the camera's view volume along the z-axis
   so that the transformed view volume will be centered on the
   negative z-axis.

     [ 1  0  -(r+l)/(2*near)  0 ]
     [ 0  1  -(t+b)/(2*near)  0 ]
     [ 0  0        1          0 ]
     [ 0  0        0          1 ]

   This matrix scales the skewed view volume so that it will
   be 2 units wide and 2 units tall at the view plane z = -1.

     [ 2*near/(r-l)       0       0  0 ]
     [       0      2*near/(t-b)  0  0 ]
     [       0            0       1  0 ]
     [       0            0       0  1 ]

   The matrix product looks like this.

     [ 1  0  -(r+l)/(2*near)  0 ]   [ 2*near/(r-l)       0       0  0 ]
     [ 0  1  -(t+b)/(2*near)  0 ] * [       0      2*near/(t-b)  0  0 ]
     [ 0  0        1          0 ]   [       0            0       1  0 ]
     [ 0  0        0          1 ]   [       0            0       0  1 ]

         [ 2*near/(r-l)       0       -(r+l)/(r-l)  0 ]
       = [   0          2*near/(t-b)  -(t+b)/(t-b)  0 ]
         [   0                0             1       0 ]
         [   0                0             0       1 ]

   This product matrix transforms the camera's perspective view
   volume into the standard (normalized) perspective view
   volume whose intersection with the view plane, z = -1,
   has left = -1, right = +1, bottom = -1, and top = +1.
</pre>
*/
public class PerspectiveNormalizeMatrix extends Matrix
{
   /**
      Construct the matrix that transforms from the camera's
      perspective view coordinate system to the normalized
      perspective camera coordinate system.
   */
   public PerspectiveNormalizeMatrix(double l, double r, double b, double t, double near)
   {
      super();

      Matrix m1, m2, m3;

      m1 = new Matrix(
               new Vector(     1.0,             0.0,        0.0, 0.0),
               new Vector(     0.0,             1.0,        0.0, 0.0),
               new Vector(-(r+l)/(2*near), -(t+b)/(2*near), 1.0, 0.0),
               new Vector(     0.0,             0.0,        0.0, 1.0));

      m2 = new Matrix(
               new Vector(2*near/(r-l),      0.0,     0.0, 0.0),
               new Vector(     0.0,     2*near/(t-b), 0.0, 0.0),
               new Vector(     0.0,          0.0,     1.0, 0.0),
               new Vector(     0.0,          0.0,     0.0, 1.0));

      m3 = m2.times( m1 );

      this.v1 = m3.v1;
      this.v2 = m3.v2;
      this.v3 = m3.v3;
      this.v4 = m3.v4;
   }
}
