/*

*/
package scene;
import scene.primitives.LineSegment;

import java.util.ArrayList;
import java.util.List;
import java.awt.Color;

/**
   A Primitive is something that we can build geometric
   shapes out of (a "graphics primitive").
<p>
   The most "primitive" primitive is a LineSegment.
<p>
   In the "primitive assembly" stage of the rendering
   pipeline, each "higher order" Primitive object needs
   to be able to "assemble" itself into a list of "lower
   order" LineSegment objects (for the clipping and
   rasterization stages of the rendering pipeline). Since
   this abstractly defined primitive does not yet know how
   to assemble itself, the assemble() method is an abstract
   method.
*/
public abstract class Primitive
{
   // A Primitive object is made up of vertices.
   public    List<Vertex> vertexList = null;
 //protected List<Vertex> vertexList = null;


   /**
      Construct an empty primitive.
   */
   protected Primitive()
   {
      this.vertexList = new ArrayList<Vertex>();
   }


   /**
      Construct a Primitive from an array of vertices.

      @param vArray array of vertices to put into the new Primitive
   */
   protected Primitive(Vertex... vArray)
   {
      this();

      for (Vertex v : vArray)
      {
         this.vertexList.add( v );
      }
   }


   /**
      Construct a Primitive from a List of vertices.

      @param vList list of vertices to put into the new Primitive
   */
   protected Primitive(List<Vertex> vList)
   {
      this.vertexList = vList;
/*
      this();

      for (Vertex v : vList)
      {
         this.vertexList.add( v );
      }
*/
   }


   /**
      Add a vertex at the end of the vertex list.

      @param v Vertex object to add to this Primitive
   */
   public void addVertex(Vertex v)
   {
      this.vertexList.add( v );
   }


   /**
      Give this primitive a uniform color.

      @param c Color for this Primitive
   */
   public void setColor(Color c)
   {
      for (Vertex v : this.vertexList)
      {
         v.setColor(c);
      }
   }

   /**
      Give this primitive a uniform, but randomly chosen, color.
   */
   public void setColorRandom()
   {
      vertexList.get(0).setColorRandom();
      Color c = vertexList.get(0).getColor();
      for (Vertex v : this.vertexList)
      {
         v.setColor(c);
      }
   }


   /**
      Give each vertex of this primitive a randomly chosen color.
   */
   public void setColorRainbow()
   {
      for (Vertex v : this.vertexList)
      {
         v.setColorRandom();
      }
   }


   /**
      Make a deep copy of this primitive.

      @return an object that is a deep copy of this Primitive object
   */
   public abstract Primitive makeCopy();


   /**
      Assemble this primitive into a list of LineSegment primitives.

      This is needed in the primitive assembly stage of the rendering pipeline.

      @return a List of the LineSegment objects that represent this Primitive
   */
   public abstract List<LineSegment> assemble();


   /**
      For debugging.

      @return String representation of this Primitive object
   */
   public abstract String toString();
}
