/*

*/
package scene.models3;
import scene.*;
import scene.primitives.*;

/**
   This is a right, triangular prism with length 2h, side length 1, centered
   at the origin, and with its central axis on the y-axis.
   http://en.wikipedia.org/wiki/Prism_(geometry)

   Attached to each of the two triangular ends of the prism is a tetrahedron
   of height h2. So the total height is 2*(h + h2).
*/
public class PrismModel extends Model
{
   /**
      A prism of length 1 with a regular tetrahedron at the top and bottom.
   */
   public PrismModel( )
   {
      this(0, 1, (float)(Math.sqrt(2)/Math.sqrt(3)) );
   }


   /**
      A prism of length h with a regular tetrahedron at the top and bottom.
   */
   public PrismModel(double h)
   {
      this(0, h, (float)(Math.sqrt(2)/Math.sqrt(3)) );
   }


   /**
      A prism of length h with a regular tetrahedron at the top and bottom
      and n "lines of latitude" around the body of the prism.
   */
   public PrismModel(int n, double h)
   {
      this(n, h, (float)(Math.sqrt(2)/Math.sqrt(3)) );
   }


   /**
      A prism of length h with a tetrahedron at the top and bottom,
      where the face-edge-face angle of the tetrahedron is theta
      (with theta in radians!) and n "lines of latitude" around
      the body of the prism.

      If theta = 0, then there are no tetrahedrons at the ends of the prism.
      If theta = arctan(2*sqrt(2)), then the tetrahedrons are regular.
   */
   public PrismModel(int n, double h, double theta)
   {
      this(n, h, (float)(Math.tan(theta)/(2*Math.sqrt(3))) );
   }


   /**
      A prism of length h with a tetrahedron of height h2 at the top and bottom
      and n "lines of latitude" around the body of the prism.

      NOTE: h2 is a float so that we can distinguish this (private) constructor
            from the (public) previous one.
   */
   private PrismModel(int n, double h, float h2)
   {
      super();

      // Create the prism's geometry.
      double sqrt3 = Math.sqrt(3.0);
      double sqrt2 = Math.sqrt(2.0);
      Vertex[] v = new Vertex[8];

      v[0] = new Vertex( 1.0/sqrt3,  h,  0);    // three vertices around the top
      v[1] = new Vertex(-0.5/sqrt3,  h,  0.5);
      v[2] = new Vertex(-0.5/sqrt3,  h, -0.5);

      v[3] = new Vertex( 1.0/sqrt3, -h,  0);    // three vertices around the bottom
      v[4] = new Vertex(-0.5/sqrt3, -h,  0.5);
      v[5] = new Vertex(-0.5/sqrt3, -h, -0.5);

      v[6] = new Vertex( 0,  h+h2, 0);  // vertex at the top
      v[7] = new Vertex( 0, -h-h2, 0);  // vertex at the bottom

      if (n == 0)
      {
         // Create three square faces.
         Primitive face = new Face();
         face.addVertex( v[0] );
         face.addVertex( v[1] );
         face.addVertex( v[4] );
         face.addVertex( v[3] );
         this.addPrimitive( face );

         face = new Face();
         face.addVertex( v[2] );
         face.addVertex( v[0] );
         face.addVertex( v[3] );
         face.addVertex( v[5] );
         this.addPrimitive( face );

         face = new Face();
         face.addVertex( v[1] );
         face.addVertex( v[2] );
         face.addVertex( v[5] );
         face.addVertex( v[4] );
         this.addPrimitive( face );
      }
      else // Create n lines of "latitude" around the prism.
      {
         double delta_y = 2.0*h/(n+1);

         Vertex[][] v2 = new Vertex[n+2][3];

         v2[0][0] = v[0];
         v2[0][1] = v[1];
         v2[0][2] = v[2];

         v2[n+1][0] = v[3];
         v2[n+1][1] = v[4];
         v2[n+1][2] = v[5];

         for (int j = 0; j <= n; j++)
         {
            double y = -h + (j+0) * delta_y;  // j+0 or j+1 ??????
            v2[j+1][0] = new Vertex( 1.0/sqrt3,  y,  0);
            v2[j+1][1] = new Vertex(-0.5/sqrt3,  y,  0.5);
            v2[j+1][2] = new Vertex(-0.5/sqrt3,  y, -0.5);

            Primitive face = new Face();
            face.addVertex( v2[j+1][0] );
            face.addVertex( v2[j+1][1] );
            face.addVertex( v2[j][1] );
            face.addVertex( v2[j][0] );
            this.addPrimitive( face );

            face = new Face();
            face.addVertex( v2[j+1][1] );
            face.addVertex( v2[j+1][2] );
            face.addVertex( v2[j][2] );
            face.addVertex( v2[j][1] );
            this.addPrimitive( face );

            face = new Face();
            face.addVertex( v2[j+1][2] );
            face.addVertex( v2[j+1][0] );
            face.addVertex( v2[j][0] );
            face.addVertex( v2[j][2] );
            this.addPrimitive( face );
         }
      }

      // Create tree triangular faces at the top.
      Primitive face = new Face();
      face.addVertex( v[6] );
      face.addVertex( v[1] );
      face.addVertex( v[0] );
      this.addPrimitive( face );

      face = new Face();
      face.addVertex( v[6] );
      face.addVertex( v[2] );
      face.addVertex( v[1] );
      this.addPrimitive( face );

      face = new Face();
      face.addVertex( v[6] );
      face.addVertex( v[0] );
      face.addVertex( v[2] );
      this.addPrimitive( face );

      // Create tree triangular faces at the bottom.
      face = new Face();
      face.addVertex( v[7] );
      face.addVertex( v[3] );
      face.addVertex( v[4] );
      this.addPrimitive( face );

      face = new Face();
      face.addVertex( v[7] );
      face.addVertex( v[4] );
      face.addVertex( v[5] );
      this.addPrimitive( face );

      face = new Face();
      face.addVertex( v[7] );
      face.addVertex( v[5] );
      face.addVertex( v[3] );
      this.addPrimitive( face );
   }
}