/*

*/
package scene.primitives;
import scene.*;

import java.util.LinkedList;
import java.util.ArrayList;
import java.util.List;

/**
   A Face is like a LineLoop, but it is assumed to be "flat" and "convex".
<p>
   Flat means that the vertices in the Face all lie in the same plane.
<p>
   Convex means that a line segment joining any two vertices of the face
   must be completely contained inside of the face.
<p>
   The combination of flat and convex means that we can give a Face an
   orientation and we can determine if the Face if "front facing" or
   "back facing".
<p>
   A Face knows how to assemble itself into a list of
   LineSegment objects, so we can implementation the
   Primitive interface.
*/
public class Face extends Primitive
{
   /**
      Construct a default (empty) Face.
   */
   public Face()
   {
      super();
   }


   /**
      Construct a Face from an array of vertices.

      @param vArray array of vertices to put into the new Face
   */
   public Face(Vertex... vArray)
   {
      super(vArray);
   }


   /**
      Construct a Face from a List of vertices.

      @param vList list of vertices to put into the new Face
   */
   public Face(List<Vertex> vList)
   {
      super(vList);
   }


   /**
      Implement the makeCopy() method from the abstract base class.
      <p>
      Make a deep copy of this Face primitive.

      @return a Face object that is a deep copy of this LineFan
   */
   public Primitive makeCopy()
   {
      List<Vertex> resultList = new ArrayList<Vertex>( this.vertexList.size() );
      for (Vertex v : this.vertexList)
      {
         resultList.add( new Vertex(v) );
      }
      return new Face( resultList );
   }


   /**
      Implement the assemble() method from the abstract base class.
      <p>
      We know how to assemble a Face into a
      list of LineSegment objects.
      <p>
      If there are n vertices in the Face,
      then we need to create n LineSegment objects.

      @return a List of the LineSegment objects that represent this Face
   */
   public List<LineSegment> assemble()
   {
    //List<LineSegment> result = new LinkedList<LineSegment>();
      List<LineSegment> result = new ArrayList<LineSegment>();

      for (int i = 0; i < this.vertexList.size() - 1; i++)
      {
         LineSegment ls = new LineSegment(new Vertex(this.vertexList.get(i+0)),
                                          new Vertex(this.vertexList.get(i+1)));
         result.add( ls );
      }

      // close the face into a loop
      result.add( new LineSegment(new Vertex(this.vertexList.get(this.vertexList.size() - 1)),
                                  new Vertex(this.vertexList.get(0))) );

      return result;
   }


   /**
      For debugging.
   */
   public String toString()
   {
      String result = "Face:\n";
      for (Vertex v : this.vertexList)
      {
         result += v.toString();
      }
      return result;
   }
}