/*

*/
package scene.primitives;
import scene.*;

import java.util.LinkedList;
import java.util.ArrayList;
import java.util.List;

/**
   A LineFan is a collection of vertices that represent
   line segments radiating out from the first given vertex.
   A LineFan with n vertices has n-1 line segments, and
   every line segment has the first given vertex as its
   initial vertex.
<p>
   A LineFan knows how to assemble itself into a list
   of LineSegment objects, so we can implementation the
   Primitive interface.
*/
public class LineFan extends Primitive
{
   /**
      Construct a default (empty) LineFan.
   */
   public LineFan()
   {
      super();
   }


   /**
      Construct a LineFan from an array of vertices.

      @param vArray array of vertices to put into the new LineFan
   */
   public LineFan(Vertex... vArray)
   {
      super(vArray);
   }


   /**
      Construct a LineFan from a List of vertices.

      @param vList list of vertices to put into the new LineFan
   */
   public LineFan(List<Vertex> vList)
   {
      super(vList);
   }


   /**
      Implement the makeCopy() method from the abstract base class.
      <p>
      Make a deep copy of this LineFan primitive.

      @return a LineFan object that is a deep copy of this LineFan
   */
   public Primitive makeCopy()
   {
      List<Vertex> resultList = new ArrayList<Vertex>( this.vertexList.size() );
      for (Vertex v : this.vertexList)
      {
         resultList.add( new Vertex(v) );
      }
      return new LineFan( resultList );
   }


   /**
      Implement the assemble() method from the abstract base class.
      <p>
      We know how to assemble a line fan into a
      list of LineSegment objects.
      <p>
      If there are n vertices in the line fan,
      then we need to create n-1 LineSegment objects.

      @return a List of the LineSegment objects that represent this LineFan
   */
   public List<LineSegment> assemble()
   {
    //List<LineSegment> result = new LinkedList<LineSegment>();
      List<LineSegment> result = new ArrayList<LineSegment>();

      for (int i = 1; i < this.vertexList.size(); i++)
      {
         LineSegment ls = new LineSegment(new Vertex(this.vertexList.get(0)),
                                          new Vertex(this.vertexList.get(i)));
         result.add( ls );
      }

      return result;
   }


   /**
      For debugging.

      @return String representation of this LineFan object
   */
   public String toString()
   {
      String result = "Line Fan:\n";
      for (Vertex v : this.vertexList)
      {
         result += v.toString();
      }
      return result;
   }
}
