/*

*/
package scene.primitives;
import scene.*;

import java.util.LinkedList;
import java.util.ArrayList;
import java.util.List;

/**
   A LineLoop is like a LineStrip, but there is one
   additional line segment connecting the last given
   vertex in the first vertex. So a LineLoop represents
   a closed curve. A LineLoop with n vertices is a curve
   with n line segments.
<p>
   A LineLoop knows how to assemble itself into a list
   of LineSegment objects, so we can implementation the
   Primitive interface.
*/
public class LineLoop extends Primitive
{
   /**
      Construct a default (empty) LineLoop.
   */
   public LineLoop()
   {
      super();
   }


   /**
      Construct a LineLoop from an array of vertices.

      @param vArray array of vertices to put into the new LineLoop
   */
   public LineLoop(Vertex... vArray)
   {
      super(vArray);
   }


   /**
      Construct a LineLoop from a List of vertices.

      @param vList list of vertices to put into the new LineLoop
   */
   public LineLoop(List<Vertex> vList)
   {
      super(vList);
   }


   /**
      Implement the makeCopy() method from the abstract base class.
      <p>
      Make a deep copy of this LineLoop primitive.

      @return a LineLoop object that is a deep copy of this LineLoop
   */
   public Primitive makeCopy()
   {
      List<Vertex> resultList = new ArrayList<Vertex>( this.vertexList.size() );
      for (Vertex v : this.vertexList)
      {
         resultList.add( new Vertex(v) );
      }
      return new LineLoop( resultList );
   }


   /**
      Implement the assemble() method from the abstract base class.
      <p>
      We know how to assemble a line loop into a
      list of LineSegment objects.
      <p>
      If there are n vertices in the line loop,
      then we need to create n LineSegment objects.

      @return a List of the LineSegment objects that represent this LineLoop
   */
   public List<LineSegment> assemble()
   {
    //List<LineSegment> result = new LinkedList<LineSegment>();
      List<LineSegment> result = new ArrayList<LineSegment>();

      for (int i = 0; i < this.vertexList.size() - 1; i++)
      {
         LineSegment ls = new LineSegment(new Vertex(this.vertexList.get(i+0)),
                                          new Vertex(this.vertexList.get(i+1)));
         result.add( ls );
      }

      // close the line strip into a line loop
      result.add( new LineSegment(new Vertex(this.vertexList.get(this.vertexList.size() - 1)),
                                  new Vertex(this.vertexList.get(0))) );

      return result;
   }


   /**
      For debugging.

      @return String representation of this LineLoop object
   */
   public String toString()
   {
      String result = "Line Loop:\n";
      for (Vertex v : this.vertexList)
      {
         result += v.toString();
      }
      return result;
   }
}
