/*

*/
package scene.primitives;
import scene.*;

import java.util.LinkedList;
import java.util.ArrayList;
import java.util.List;

/**
   A LineStrip is a collection of vertices that
   represent consecutively connected line segments.
   A LineStrip represents a curve. If the LineStrip
   contains n vertices, then the curve is made up of
   n-1 line segments.
<p>
   A LineStrip knows how to assemble itself into a list
   of LineSegment objects, so we can implementation the
   Primitive interface.
*/
public class LineStrip extends Primitive
{
   /**
      Construct a default (empty) LineStrip.
   */
   public LineStrip()
   {
      super();
   }


   /**
      Construct a LineStrip from an array of vertices.

      @param vArray array of vertices to put into the new LineStrip
   */
   public LineStrip(Vertex... vArray)
   {
      super(vArray);
   }


   /**
      Construct a LineStrip from a List of vertices.

      @param vList list of vertices to put into the new LineStrip
   */
   public LineStrip(List<Vertex> vList)
   {
      super(vList);
   }


   /**
      Implement the makeCopy() method from the abstract base class.
      <p>
      Make a deep copy of this LineStrip primitive.

      @return a LineStrip object that is a deep copy of this LineStrip
   */
   public Primitive makeCopy()
   {
      List<Vertex> resultList = new ArrayList<Vertex>( this.vertexList.size() );
      for (Vertex v : this.vertexList)
      {
         resultList.add( new Vertex(v) );
      }
      return new LineStrip( resultList );
   }


   /**
      Implement the assemble() method from the abstract base class.
      <p>
      We know how to assemble a line strip into a
      list of LineSegment objects.
      <p>
      If there are n vertices in the line strip,
      then we need to create n-1 LineSegment objects.

      @return a List of the LineSegment objects that represent this LineStrip
   */
   public List<LineSegment> assemble()
   {
    //List<LineSegment> result = new LinkedList<LineSegment>();
      List<LineSegment> result = new ArrayList<LineSegment>();

      for (int i = 0; i < this.vertexList.size() - 1; i++)
      {
         LineSegment ls = new LineSegment(new Vertex(this.vertexList.get(i+0)),
                                          new Vertex(this.vertexList.get(i+1)));
         result.add( ls );
      }

      return result;
   }


   /**
      For debugging.

      @return String representation of this LineStrip object
   */
   public String toString()
   {
      String result = "Line Strip:\n";
      for (Vertex v : this.vertexList)
      {
         result += v.toString();
      }
      return result;
   }
}
